DIGITALTWIN_ON_FHIR_SYSTEM = "https://www.auckland.ac.nz/en/abi.html"

SNOMEDCT = {
    "BREAST": {
        "system": "http://snomed.info/sct",
        "code": "76752008",
        "display": "Breast"
    }
}




from typing import Optional, Literal, List

InputResourceType = ["Observation", "ImagingStudy", "DocumentReference", ""]
OutputResourceType = ["DiagnosticReport", *InputResourceType]


class WorkflowGoal:

    def __init__(self, description):
        self.description = description if isinstance(description, str) else str(description)

    def get(self):
        goal = {
            "description": self.description
        }
        return {k: v for k, v in goal.items() if v not in ("", None)}


class WorkflowActionInput:

    def __init__(self, display: str, resource_type: Literal["Observation", "ImagingStudy", "DocumentReference", ""] = ""
                 ):
        if not isinstance(display, str):
            raise ValueError(f"display should be a string")
        self.resource_type = resource_type
        self.display = display

    def set_resource_type(self, resource_type: Literal["Observation", "ImagingStudy", "DocumentReference"]):
        if resource_type not in InputResourceType:
            raise ValueError(f"resource_type should be one of {InputResourceType}")
        self.resource_type = resource_type
        return self

    def get(self):
        action_input = {
            "resource_type": self.resource_type,
            "display": self.display,
        }
        return {k: v for k, v in action_input.items() if v not in ("", None)}


class WorkflowActionOutput:
    def __init__(self, display: str,
                 resource_type: Literal["Observation", "ImagingStudy", "DocumentReference", ""] = "",
                 code: Optional = None,
                 system: Optional = None,
                 unit: Optional = None):
        if not isinstance(display, str):
            raise ValueError(f"display should be a string")
        self.resource_type = resource_type
        self.code = code
        self.system = system
        self.display = display
        self.unit = unit

    def set_resource_type(self, resource_type: Literal[
        "Observation", "ImagingStudy", "DocumentReference", "DiagnosticReport"]):
        print(OutputResourceType)
        if resource_type not in OutputResourceType:
            raise ValueError(f"resource_type should be one of {OutputResourceType}")
        self.resource_type = resource_type
        return self

    def set_code(self, code: str):
        if not isinstance(code, str):
            raise ValueError(f"code should be a string")
        self.code = code
        return self

    def set_system(self, system: str):
        if not isinstance(system, str):
            raise ValueError(f"system should be a string")
        self.system = system
        return self

    def set_unit(self, unit: str):
        if not isinstance(unit, str):
            raise ValueError(f"unit should be a string")
        self.unit = unit
        return self

    def get(self):
        action_output = {
            "resource_type": self.resource_type,
            "display": self.display,
            "code": self.code,
            "system": self.system,
            "unit": self.unit
        }
        if self.resource_type == "Observation":
            action_output["code"] = self.code
            action_output["system"] = self.system
            action_output["unit"] = self.unit
        return {k: v for k, v in action_output.items() if v not in ("", None)}


class WorkflowAction:
    def __init__(self, title: str, description: str, action_input: List[WorkflowActionInput],
                 action_output: List[WorkflowActionOutput], related_tool_uuid: str = ""):
        if not isinstance(title, str):
            raise ValueError(f"title should be a string")
        elif not isinstance(action_input, list):
            raise ValueError(f"action_input should be a WorkflowActionInput list")
        elif not isinstance(action_output, list):
            raise ValueError(f"action_output should be a WorkflowActionOutput list")

        self.title = title
        self.description = description if isinstance(description, str) else "No description available."
        self.related_tool_uuid = related_tool_uuid
        self.action_input = action_input
        self.action_output = action_output

    def set_related_tool_uuid(self, related_tool_uuid: str):
        if not isinstance(related_tool_uuid, str):
            raise ValueError(f"related_tool_uuid should be a string")
        self.related_tool_uuid = related_tool_uuid
        return self

    def annotate_input(self) -> list[WorkflowActionInput]:
        return self.action_input

    def annotate_output(self) -> list[WorkflowActionOutput]:
        return self.action_output

    def get(self):
        action = {
            "title": self.title,
            "description": self.description,
            "related_tool_uuid": self.related_tool_uuid,
            "input": [action_input.get() for action_input in self.action_input if
                      isinstance(action_input, WorkflowActionInput)],
            "output": [action_output.get() for action_output in self.action_output if
                       isinstance(action_output, WorkflowActionOutput)]
        }
        return {k: v for k, v in action.items() if v not in ("", None)}



from abc import ABC
from .abstract_annotator import AbstractAnnotator
from pprint import pprint
from fhir_cda.utils import ordered_load
from fhir_cda.ehr.workflow import (WorkflowGoal, WorkflowActionInput, WorkflowActionOutput, WorkflowAction)
from typing import List, Optional, Union
import copy


class WorkflowAnnotator(AbstractAnnotator, ABC):
    def __init__(self, dataset_path):
        super().__init__(dataset_path, "workflow")
        self.cwl_content = {}
        self.elements = {}
        self._analysis_workflow()

    def _analysis_workflow(self):
        primary_folder = self._root / "primary"

        workflow_paths = list(primary_folder.glob("*.cwl"))
        if not workflow_paths:
            self._descriptions = {}
            raise Exception("No workflow cwl found")

        workflow_path = workflow_paths[0]
        with open(workflow_path, 'r') as file:
            self.cwl_content = ordered_load(file)

        self.elements["workflow"] = {
            "uuid": "",
            "name": self._root.name,
            "title": self.cwl_content.get("class"),
            "version": self.cwl_content.get("cwlVersion"),
            "description": "",
            "purpose": "",
            "usage": "",
            "author": "",
            "goal": [],
            "action": []
        }

        self._generate_action_goal(self.cwl_content["steps"])

    def _generate_action_goal(self, steps):
        workflow_goals = []
        for idx, step in enumerate(steps.keys()):
            inputs = steps[step]["run"]["inputs"]
            outputs = steps[step]["run"]["outputs"]
            workflow_action_inputs = []
            workflow_action_outputs = []

            for i, name in enumerate(inputs.keys()):
                workflow_action_inputs.append(WorkflowActionInput(
                    display=name
                ))
            for j, name in enumerate(outputs.keys()):
                workflow_action_outputs.append(WorkflowActionOutput(
                    display=name
                ))
                workflow_goals.append(WorkflowGoal(description=name))

            self.elements["workflow"]["action"].append(WorkflowAction(
                title=step,
                description=f'step {idx + 1}',
                action_input=workflow_action_inputs,
                action_output=workflow_action_outputs
            ))
        self.update_goals(workflow_goals)

    def _convert_elements_to_descriptions(self):
        self._descriptions["workflow"] = copy.deepcopy(self.elements["workflow"])
        self._descriptions["workflow"]["goal"] = []
        self._descriptions["workflow"]["action"] = []
        for g in self.elements["workflow"].get("goal", []):
            self._descriptions["workflow"]["goal"].append(g.get())
        for a in self.elements["workflow"].get("action", []):
            self._descriptions["workflow"]["action"].append(a.get())

    def update_workflow_field(self, field, value):
        if field not in self.elements.get("workflow"):
            raise ValueError(f"field {field} is not in descriptions['workflow_tool']")
        else:
            self.elements["workflow"][field] = value
        return self

    def update_name(self, name: str):
        self.elements["workflow"]["name"] = name
        return self

    def update_goals(self, goals: List[WorkflowGoal]):
        if not isinstance(goals, List):
            return self
        else:
            self.elements["workflow"]["goal"] = [g for g in goals if isinstance(g, WorkflowGoal)]
            return self

    def add_goal(self, goal: WorkflowGoal):
        if isinstance(goal, WorkflowGoal):
            self.elements["workflow"]["goal"].append(goal)
        else:
            raise ValueError(f"Goal {goal} is not a WorkflowGoal")
        return self

    def update_uuid(self, uuid: str):
        if isinstance(uuid, str):
            self.elements["workflow"]["uuid"] = uuid
        else:
            raise ValueError(f"UUID {uuid} is not a string")
        return self

    def update_title(self, title: str):
        if isinstance(title, str):
            self.elements["workflow"]["title"] = title
        else:
            raise ValueError(f"Title {title} is not a string")
        return self

    def update_version(self, version: str):
        if isinstance(version, str):
            self.elements["workflow"]["version"] = version
        else:
            raise ValueError(f"Version {version} is not a string")
        return self

    def update_description(self, description):
        if isinstance(description, str):
            self.elements["workflow"]["description"] = description
        else:
            raise ValueError(f"Description {description} is not a string")
        return self

    def update_purpose(self, purpose: str):
        if isinstance(purpose, str):
            self.elements["workflow"]["purpose"] = purpose
        else:
            raise ValueError(f"Purpose {purpose} is not a string")
        return self

    def update_usage(self, usage: str):
        if isinstance(usage, str):
            self.elements["workflow"]["usage"] = usage
        else:
            raise ValueError(f"Usage {usage} is not a string")
        return self

    def update_author(self, author: str):
        if isinstance(author, str):
            self.elements["workflow"]["author"] = author
        else:
            raise ValueError(f"Author {author} is not a string")
        return self

    def annotate_action(self, step: Optional[int] = None) -> Union[WorkflowAction, list[WorkflowAction]]:
        actions = self.elements["workflow"]["action"]
        if isinstance(step, int) and 0 <= step < len(actions):
            print(f"Get step {step} action {actions[step].get()}")
            return actions[step]
        else:
            print(f"Get total {len(actions)} action, you can use .get() method to get action details")
            return actions

    def get_descriptions(self):
        self._convert_elements_to_descriptions()
        return self._descriptions

    def save(self, path=None):
        self._convert_elements_to_descriptions()
        super().save(path)



from typing import Optional, List
from .elements import ObservationValue, ImagingStudySeries, ImagingStudySeriesInstance, DocumentAttachment
import pydicom
from pydicom.uid import UID
from fhir_cda.terms import SNOMEDCT
from concurrent.futures import ThreadPoolExecutor
import os
from pathlib import Path


class ObservationMeasurement:
    def __init__(self, value: Optional[ObservationValue] = None, code: Optional[str] = "",
                 code_system="http://loinc.org",
                 unit: Optional[str] = None,
                 display: Optional[str] = None, uuid: Optional[str] = ""):

        if value is not None and not isinstance(value, ObservationValue):
            raise ValueError(f"value={value} is not an ObservationValue type")
        elif not isinstance(code, str):
            raise ValueError(f"code={code} is not an instance of type str")
        elif unit is not None and not isinstance(unit, str):
            raise ValueError(f"unit={unit} is not an instance of type str")
        elif display is not None and not isinstance(display, str):
            raise ValueError(f"display={display} is not an instance of type str")
        elif not isinstance(uuid, str):
            raise ValueError(f"uuid={uuid} is not an instance of type str")

        self.measurement_type = "ObservationMeasurement"
        self.uuid = uuid
        self.value = value
        self.code = code
        self.unit = unit
        self.code_system = code_system
        self.display = display

    def __repr__(self):
        return (
            f"ObservationMeasurement(uuid={self.uuid},value={self.value}, code='{self.code}', value_system='{self.code_system}')")

    def set(self, item):
        self.uuid = item.get("uuid", "")
        self.code = item.get("code", "")
        self.display = item.get("display", "")
        self.code_system = item.get("codeSystem", "")
        self.unit = item.get("unit", "")
        value = item.get("value", None)
        self.value = ObservationValue().set(value) if value is not None else None
        return self

    def set_uuid(self, uuid: str):
        self.uuid = uuid
        return self

    def set_value(self, value: ObservationValue):
        self.value = value
        return self

    def set_unit(self, unit: str):
        self.unit = unit
        return self

    def set_code_system(self, code_system: str):
        self.code_system = code_system
        return self

    def set_display(self, display: str):
        self.display = display
        return self

    def set_code(self, code: str):
        self.code = code
        return self

    def get(self):
        measurement = {
            "resourceType": "Observation",
            "uuid": self.uuid,
            "value": self.value.get(),
            "code": self.code,
            "codeSystem": self.code_system,
            "unit": self.unit,
            "display": self.display if isinstance(self.display, str) else ""
        }
        return {k: v for k, v in measurement.items() if v is not None}



class DocumentReferenceMeasurement:
    def __init__(self, attachments: Optional[List[DocumentAttachment]] = None, title: Optional[str] = "",
                 uuid: Optional[str] = ""):
        if not (isinstance(attachments, list) and all(isinstance(item, DocumentAttachment) for item in attachments)):
            raise ValueError(f"attachments={attachments} is not an instance of type list")
        if not isinstance(title, str):
            raise ValueError(f"title={title} is not an instance of type str")
        elif not isinstance(uuid, str):
            raise ValueError(f"uuid={uuid} is not an instance of type str")

        self.measurement_type = "DocumentReferenceMeasurement"
        self.uuid = uuid
        self.attachments = attachments
        self.title = title

    def __repr__(self):
        return (
            f"DocumentReferenceMeasurement(uuid={self.uuid}, title={self.title})")

    def set(self, item):
        self.uuid = item.get("uuid", "")
        self.title = item.get("title", "")
        self.attachments = item.get("attachments", [])
        return self

    def set_uuid(self, uuid: str):
        self.uuid = uuid
        return self

    def set_attachments(self, attachments: List[DocumentAttachment]):
        if not isinstance(attachments, list):
            raise ValueError(f"attachments={attachments} is not an instance of type list")
        elif not all(isinstance(item, DocumentAttachment) for item in attachments):
            raise ValueError(f"attachments={attachments}: some attachments are not an DocumentAttachment type")
        self.attachments = attachments
        return self

    def set_title(self, title: str):
        self.title = title
        return self

    def get(self):
        measurement = {
            "resourceType": "DocumentReference",
            "uuid": self.uuid,
            "title": self.title,
            "attachments": [a.get() for a in self.attachments if isinstance(a, DocumentAttachment)]
        }
        return measurement


class ImagingStudyMeasurement:
    """
    uuid: Imaging Study Identifier, should be generated by digitaltwins-on-fhir, fhir-cda, digitaltwins platform or user.
    endpoint_uuid (ImagingStudy): Imaging Study endpoint Identifier, should be generated by digitaltwins-on-fhir and fhir-cda.
    endpoint_uuid (ImagingStudySeries): Imaging Study series endpoint Identifier, can be dataset's sample uuid which generated by digitaltwins platform or user.

    sample_details: str[] or Path[]
    """

    def __init__(self, uuid: str = "", sample_details: list = None, endpoint_url: str = "", description: str = ""):
        if sample_details is not None and not isinstance(sample_details, list):
            raise ValueError(f"samples={sample_details} is not an instance of type list")
        if not isinstance(uuid, str):
            raise ValueError(f"uuid={uuid} is not an instance of type str")
        if description is not None and not isinstance(description, str):
            raise ValueError(f"description={description} is not an instance of type str")
        if endpoint_url is not None and not isinstance(endpoint_url, str):
            raise ValueError(f"endpoint_url={endpoint_url} is not an instance of type str")

        self.measurement_type = "ImagingStudyMeasurement"
        self.uuid = uuid
        self.endpoint_url = endpoint_url
        self.description = description
        self.series = []
        self._samples = []

        if sample_details is not None:
            if len(sample_details) == 0:
                raise ValueError(f"sample_paths={sample_details} should have at least one instance.")
            for sample in sample_details:
                if isinstance(sample, str):
                    self._samples.append({
                        "uuid": "",
                        "path": Path(sample)
                    })
                elif isinstance(sample, Path):
                    self._samples.append({
                        "uuid": "",
                        "path": sample
                    })
                elif self._is_valid_sample_dict(sample):
                    if isinstance(sample["path"], str):
                        sample["path"] = Path(sample["path"])
                    self._samples.append(sample)
                else:
                    raise ValueError(f"sample type {type(sample)} is not a valid type.")
            self._generate_imaging_study()

    def __repr__(self):
        return (
            f"ImagingStudyMeasurement(uuid={self.uuid},endpointUrl={self.endpoint_url}, description={self.description}, series={len(self.series)})")

    @staticmethod
    def _is_valid_sample_dict(d):
        return (
                isinstance(d, dict)
                and "uuid" in d and isinstance(d["uuid"], str)
                and "path" in d and (isinstance(d["path"], Path) or isinstance(d["path"], str))
        )

    def set(self, item):
        self.uuid = item.get("uuid", "")
        self.endpoint_url = item.get("endpoint_url", "")
        self.description = item.get("description", "")
        if item.get("series", None) is not None:
            self.series = [ImagingStudySeries().set(s) for s in item.get("series")] if isinstance(item.get("series"),
                                                                                                  list) else []
        else:
            self.series = []

        return self

    def set_uuid(self, uuid):
        self.uuid = uuid
        return self

    def set_endpoint_url(self, endpoint_url):
        self.endpoint_url = endpoint_url
        return self

    def set_description(self, description):
        self.description = description
        return self

    def get(self):
        imaging_study_measurement = {
            "resourceType": "ImagingStudy",
            "uuid": self.uuid if isinstance(self.uuid, str) else "",
            "endpointUrl": self.endpoint_url if isinstance(self.endpoint_url, str) else "",
            "description": self.description if isinstance(self.description, str) else "",
            "series": [s.get() for s in self.series if isinstance(s, ImagingStudySeries)] if isinstance(self.series,
                                                                                                        list) else [],
        }
        return imaging_study_measurement

    def get_series(self):
        return self.series

    def _generate_imaging_study(self):

        primary_folder = self._samples[0]["path"].parent.parent
        patients_dirs = [x for x in primary_folder.iterdir() if x.is_dir()]

        if len(patients_dirs) < 5:
            for sam in self._samples:
                s = self._read_sam(sam)
                if s is not None:
                    self.series.append(s)
        else:
            self.series.extend(self._analysis_dicom_samples_worker(self._samples))

    def _read_sam(self, sam):
        try:
            dcm_files = list(sam["path"].glob("*.dcm"))
            nrrd_files = list(sam["path"].glob("*.nrrd"))
            nii_files = list(sam["path"].glob("*.nii.gz"))
            if len(dcm_files) < 1 and len(nrrd_files) < 1 and len(nii_files) < 1:
                return
            if (len(dcm_files) > 0 and len(nrrd_files) > 0) or (len(dcm_files) > 0 and len(nii_files) > 0) or (
                    len(nrrd_files) > 0 and len(nii_files) > 0):
                raise ValueError(
                    "dataset format error: Detected dcm, nii.gz and nrrd files under the same sample folder.")

            if len(dcm_files) >= 1:
                s_dicom_file = pydicom.dcmread(dcm_files[0])
                body_part_examined = s_dicom_file.get((0x0018, 0x0015), None)
                body_site = SNOMEDCT.get(body_part_examined.value.upper(),
                                         None) if body_part_examined is not None else None

                suid = s_dicom_file.get((0x0020, 0x000e), None)
                s = ImagingStudySeries(uid=suid.value if suid is not None else "",
                                       endpoint_url="",
                                       endpoint_uuid=sam["uuid"],
                                       name=sam["path"].name,
                                       number_of_instances=len(dcm_files),
                                       body_site=body_site,
                                       instances=self._analysis_dicom_sample_instances(dcm_files)
                                       )
                return s
            if len(nrrd_files) >= 1:
                s = ImagingStudySeries(uid=None,
                                       endpoint_url="",
                                       endpoint_uuid=sam["uuid"],
                                       name=sam["path"].name,
                                       number_of_instances=len(nrrd_files),
                                       instances=[]
                                       )
                return s
            if len(nii_files) >= 1:
                s = ImagingStudySeries(uid=None,
                                       endpoint_url="",
                                       endpoint_uuid=sam["uuid"],
                                       name=sam["path"].name,
                                       number_of_instances=len(nii_files),
                                       instances=[]
                                       )
                return s
        except Exception as e:
            print(f"Error reading {sam}: {e}")
            return None

    def _analysis_dicom_samples_worker(self, sams):
        samples = []
        max_workers = os.cpu_count()
        with ThreadPoolExecutor(max_workers=max_workers) as executor:
            results = executor.map(self._read_sam, sams)

        for result in results:
            if result is not None:
                samples.append(result)

        return samples

    @staticmethod
    def _analysis_dicom_sample_instances(dcms):
        instances = []
        # TODO: 14/05/2025 decision: not consider the instance level at this stage
        # for d in dcms:
        #     dcm = pydicom.dcmread(d)
        #
        #     # Get the SOP Class UID
        #     sop_class_uid = dcm.SOPClassUID
        #     # Get the SOP Class Name using the UID dictionary
        #     sop_class_name = UID(sop_class_uid).name
        #
        #     instance = ImagingStudySeriesInstance(uid=dcm[(0x0008, 0x0018)].value, sop_class_uid=sop_class_uid,
        #                                           sop_class_name=sop_class_name, number=dcm[(0x0020, 0x0013)].value)
        #
        #     # instance = {
        #     #     "uid": dcm[(0x0008, 0x0018)].value,
        #     #     "sopClassUid": sop_class_uid,
        #     #     "sopClassName": sop_class_name,
        #     #     "number": dcm[(0x0020, 0x0013)].value
        #     # }
        #    instances.append(instance)
        return instances



from pathlib import Path
from abc import ABC
import json


class AbstractAnnotator(ABC):

    def __init__(self, dataset_path, category):
        self._root = Path(dataset_path)
        self._category = category
        self._descriptions = {}

    def save(self, path=None):
        if path:
            save_path = Path(path) / f"{self._category}.json"
        else:
            save_path = self._root / f"{self._category}.json"

        with open(save_path, "w") as json_file:
            json.dump(self._descriptions, json_file, indent=4)


from abc import ABC
from .abstract_annotator import AbstractAnnotator
from typing import Literal

from fhir_cda.ehr import ObservationMeasurement, DocumentReferenceMeasurement, ImagingStudyMeasurement
import json

from ..utils import check_first_file_extension
import copy
from pathlib import Path


class MeasurementAnnotator(AbstractAnnotator, ABC):

    def __init__(self, dataset_path, mode="default"):
        """

        :param dataset_path:
        :param mode: string, "default" or "update"
        """
        super().__init__(dataset_path, "measurements")
        self._patient_paths = []
        self.elements = {}
        if mode == "update":
            self._read_measurements()
        else:
            self._analysis_dataset()

    def _read_measurements(self):
        measurements_path = self._root / "measurements.json"
        if not measurements_path.exists():
            raise ValueError("Measurements json file does not exist!")
        else:
            with open(measurements_path, "r") as f:
                self._descriptions = json.load(f)
                self._convert_descriptions_to_elements()

    def _convert_descriptions_to_elements(self):
        self.elements["dataset"] = copy.deepcopy(self._descriptions["dataset"])
        self.elements["patients"] = []
        for idx, p in enumerate(self._descriptions.get("patients", [])):
            patient_element = {
                "uuid": p.get("uuid", ""),
                "name": p.get("name", ""),
                "observations": [ObservationMeasurement().set(o) for o in p.get("observations", [])],
                "imagingStudy": [ImagingStudyMeasurement().set(i) for i in p.get("imagingStudy", [])],
                "documentReference": [DocumentReferenceMeasurement().set(d) for d in p.get("documentReference", [])],
            }
            self.elements["patients"].append(patient_element)
        self.elements["patients"] = sorted(self.elements["patients"], key=lambda x: x["name"])

    def _convert_elements_to_descriptions(self):
        self._descriptions["dataset"] = copy.deepcopy(self.elements["dataset"])
        self._descriptions["patients"] = []
        for idx, p in enumerate(self.elements.get("patients", [])):
            patient_description = {
                "uuid": p.get("uuid", ""),
                "name": p.get("name", ""),
                "observations": [observationMeasurement.get() for observationMeasurement in p.get("observations", [])],
                "imagingStudy": [imagingStudyMeasurement.get() for imagingStudyMeasurement in
                                 p.get("imagingStudy", [])],
                "documentReference": [documentReferenceMeasurement.get() for documentReferenceMeasurement in
                                      p.get("documentReference", [])]

            }
            self._descriptions["patients"].append(patient_description)
        self._descriptions["patients"] = sorted(self._descriptions["patients"], key=lambda x: x["name"])

    def _analysis_dataset(self):
        primary_folder = self._root / "primary"
        if not primary_folder.exists():
            self.elements = {}
            raise ValueError(
                'The dataset structure is not based on a SPARC SDS dataset format, please check it and try again!')
        self.elements["dataset"] = {
            "uuid": "",
            "name": self._root.name,
        }
        self.elements["patients"] = []
        self._patient_paths = [x for x in primary_folder.iterdir() if x.is_dir()]
        for p in self._patient_paths:
            patient = {
                "uuid": "",
                "name": p.name,
                "observations": [],
                "imagingStudy": [],
                "documentReference": [],
            }
            self.elements["patients"].append(patient)

    def automated_generating_imaging_study_measurement_by_scan_dataset(self):
        for path, patient in zip(self._patient_paths, self.elements["patients"]):
            patient["imagingStudy"] = self._analysis_imaging_study_samples(path)
        return self

    @staticmethod
    def _analysis_imaging_study_samples(study):
        imaging_studies = []
        if study.exists():
            sams = [x for x in study.iterdir() if x.is_dir()]
            if len(sams) < 1:
                return imaging_studies
            else:
                dcm_sams = [sam for sam in sams if check_first_file_extension(sam) == "dcm"]
                nrrd_sams = [sam for sam in sams if check_first_file_extension(sam) == "nrrd"]

            if len(dcm_sams) > 0:
                imaging_study = ImagingStudyMeasurement(sample_details=dcm_sams, description="dcm")
                if len(imaging_study.series) > 0:
                    imaging_studies.append(imaging_study)
            if len(nrrd_sams) > 0:
                imaging_study = ImagingStudyMeasurement(sample_details=nrrd_sams, description="nrrd")
                if len(imaging_study.series) > 0:
                    imaging_studies.append(imaging_study)
        return imaging_studies

    def add_measurements(self, subjects, measurements):
        self.add_measurement(subjects, measurements)
        return self

    def add_measurement(self, subjects, measurement):
        if isinstance(subjects, list) and isinstance(measurement, list):
            for s in subjects:
                self.add_measurements_by_subject(s, measurement)
        elif isinstance(subjects, list) and isinstance(measurement, (
                ObservationMeasurement, DocumentReferenceMeasurement, ImagingStudyMeasurement)):
            m = measurement
            for s in subjects:
                self.add_measurement_by_subject(s, m)
        elif isinstance(subjects, str) and isinstance(measurement, list):
            s = subjects
            for m in measurement:
                self.add_measurement_by_subject(s, m)
        elif isinstance(subjects, str) and isinstance(measurement, (
                ObservationMeasurement, DocumentReferenceMeasurement, ImagingStudyMeasurement)):
            s = subjects
            m = measurement
            self.add_measurement_by_subject(s, m)
        return self

    def add_measurements_by_subject(self, subject, measurements):
        s = subject
        for m in measurements:
            self.add_measurement_by_subject(s, m)

        return self

    def add_measurement_by_subject(self, subject, measurement):
        if not isinstance(subject, str):
            raise ValueError(f"subject={subject} is not an instance of type str")
        if not isinstance(measurement, (ObservationMeasurement, DocumentReferenceMeasurement, ImagingStudyMeasurement)):
            raise ValueError(f"measurement={measurement} is not an instance of type Measurement")
        subject_path = self._root / "primary" / subject
        if not subject_path.exists():
            raise ValueError(f"subject_path={subject_path} is not exists")

        matched_patient = self._find_matched_patient(subject)
        assert isinstance(matched_patient, dict)

        if measurement.measurement_type == "ObservationMeasurement":
            matched_patient["observations"].append(measurement)
        elif measurement.measurement_type == "DocumentReferenceMeasurement":
            matched_patient["documentReference"].append(measurement)
        elif measurement.measurement_type == "ImagingStudyMeasurement":
            matched_patient["imagingStudy"].append(measurement)
        return self

    def update_dataset(self, field, value):
        if field not in self.elements.get("dataset"):
            raise ValueError(f"field {field} is not in descriptions['dataset']")
        else:
            self.elements["dataset"][field] = value

    def update_patient(self, subject, field, value):
        matched_patient = self._find_matched_patient(subject)
        if field not in matched_patient:
            raise ValueError(f"field {field} is not in descriptions['patient']")
        elif field in ["observations", "documentReference", "imagingStudy"]:
            print(
                f"detected your want to update {subject}'s {field} measurements, now forward you to update_patient_measurements() method.")
            if field == "observations":
                self.update_patient_measurements(subject, "ObservationMeasurement")
            elif field == "documentReference":
                self.update_patient_measurements(subject, "DocumentReferenceMeasurement")
            elif field == "imagingStudy":
                self.update_patient_measurements(subject, "ImagingStudyMeasurement")

        else:
            matched_patient[field] = value
        return self

    def update_patient_measurements(self,
                                    subject: str,
                                    category: Literal[
                                        "ObservationMeasurement",
                                        "ImagingStudyMeasurement",
                                        "DocumentReferenceMeasurement"]):
        matched_patient = self._find_matched_patient(subject)

        if category == "ObservationMeasurement":
            return matched_patient["observations"]
        elif category == "ImagingStudyMeasurement":
            return matched_patient["imagingStudy"]
        elif category == "DocumentReferenceMeasurement":
            return matched_patient["documentReference"]
        else:
            raise ValueError(
                f"Unrecognized category '{category}', only accepted values are ObservationMeasurement, ImagingStudyMeasurement, and DocumentReferenceMeasurement.")

    def _find_matched_patient(self, subject):
        matched_patients = [p for p in self.elements.get("patients", []) if
                            p.get("name") == subject]
        if len(matched_patients) == 0:
            raise ValueError(f"No patients found for {subject}")
        matched_patient = matched_patients[0]
        return matched_patient

    def get_descriptions(self):
        self._convert_elements_to_descriptions()
        return self._descriptions

    def save(self, path=None):
        self._convert_elements_to_descriptions()
        super().save(path)



from .core import Annotator
from datetime import datetime

__title__ = "fhir_cda"
__version__ = "1.2.4"
__author__ = "Linkun Gao"
__license__ = "Apache-2.0"
__copyright__ = f"Copyright 2024-{datetime.now().year} ABI"

# Version synonym
VERSION = __version__